/*
 * CGI clock
 * Copyright (C) 1999 Open World Ltd
 * Written by Shevek <shevek@anarres.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include "library.h"
#include "hash.h"
#include "rope.h"
#include "cgi.h"
#include "smalloc.h"
#include "log.h"

/* Define DIR_ZONES if you don't have Linux zoneinfo */
#define DIR_ZONES		DIR_DATA "/" DEFAULT "/clock/zoneinfo/"

/* Resource directory on server for images */
#define DIR_LOCATION	DIR_RESOURCE "/clock/images/"

/* Fingers out from here on */
#define STREQ(x, y) (strcmp(x, y) == 0)

typedef
struct {
	char	*tz;
	char	 digitset;
	int		 digit;
} clockinfo_t;

void
clock_show(clockinfo_t *info)
{
	time_t		 timebuf;
	struct tm	*tm;
	int			 val;
	char		*server;
	char		*env;
	int			 len;

	server = getenv("SERVER_NAME");
	if (!server) server = DEFAULT_SERVER;

	len = 4 + strlen(info->tz);
#ifdef DIR_ZONES
	len += strlen(DIR_ZONES);
#endif
	env = smalloc(len * sizeof(char));
#ifdef DIR_ZONES
	sprintf(env, "TZ=" DIR_ZONES "%s", info->tz);
#else
	sprintf(env, "TZ=%s", info->tz);
#endif
	putenv(env);
	tzset();
	time(&timebuf);
	tm = localtime(&timebuf);

	switch(info->digit) {
		case 0: val = tm->tm_min % 10; break;	/* Minute units */
		case 1: val = tm->tm_min / 10; break;	/* Minute tens */
		case 2: val = tm->tm_hour % 10; break;	/* Hour units */
		case 3: val = tm->tm_hour / 10; break;	/* Hour tens */
		default:	val = 0;	/* Defeat compile warnings */
	}

	printf("Location: http://%s" DIR_LOCATION "%c/%c.gif" "\n\n",
					server, info->digitset, ((char)(val + '0')) );
}


/* QUERY_STRING format: Traditional CGI a=b&c=d
 * Parameters:
 * location			- Unix timezone to use
 * digitPosition	- digit to show, may be (hours|minutes)(tens|units)
 * All other parameters are ignored
 */
void
clock_info_long(clockinfo_t *info)
{
	hash_t	*cgihash;
	char	*tz;
	char	*temp;

	cgihash = cgi_hash_new();

/* Get the timezone */
	tz = hash_get(cgihash, "location");
	if (!tz) {
		log("cgiconf", PRI_CRIT, "No 'location' in QUERY_STRING for "
						"long protocol");
		exit(0);
	}
	info->tz = tz;

/* Get the digit set */
	temp = hash_get(cgihash, "case");
	if (!temp)
		info->digitset = 'B';
	else if (strlen(temp) == 1)
		info->digitset = *temp;
	else if (STREQ(temp, "default"))
		info->digitset = 'B';
	else if (STREQ(temp, "grand"))
		info->digitset = 'C';
	else if (STREQ(temp, "park"))
		info->digitset = 'D';
	else if (STREQ(temp, "oeh.default"))
		info->digitset = 'E';
	else if (STREQ(temp, "intercon.default"))
		info->digitset = 'F';
	else if (STREQ(temp, "panpac.default"))
		info->digitset = 'G';
	else {
		info->digitset = 'B';
		log("cgisys", PRI_WARNING, "Invalid digit set '%s' requested",
						temp);
	}

/* Get the digit */
	temp = hash_get(cgihash, "digitPosition");
	if (!temp) {
		log("cgiconf", PRI_CRIT, "No 'digitPosition' in QUERY_STRING "
						"for long format");
		exit(0);
	}
	if (STREQ(temp, "hourstens"))
		info->digit = 3;
	else if (STREQ(temp, "hoursunits"))
		info->digit = 2;
	else if (STREQ(temp, "minutestens"))
		info->digit = 1;
	else if (STREQ(temp, "minutesunits"))
		info->digit = 0;
	else {
		log("cgiconf", PRI_CRIT, "digitPosition parameter is invalid");
		exit(0);
	}
}

/* QUERY_STRING format:
 * %c%c%s:%s
 * %c - digitset
 * %c - digit (ASCII value mod 4 is used)
 * %s - timezone
 * :%s  - literal ':' followed by any string - optional
 */
void
clock_info_short(clockinfo_t *info)
{
	rope_t	*cgirope;
	char	*temp;
	int		 i;

	/* OK, so we allocate twice. Hell. */
	cgirope = cgi_rope_new();
	temp = rope_chars(cgirope);

	i = strcspn(temp, ":");
	if (i < 2) {
		log("cgisys", PRI_CRIT, "strlen(QUERY_STRING) < 2 for short "
						"protocol");
		exit(0);
	}
	temp[i] = 0;			/* Modify rope by reference */
	info->digitset = *temp++;
	info->digit = ((unsigned char)(*temp++)) % 4;
	info->tz = temp;				/* Still points into rope */
}

void
invoke(void)
{
	rope_t		*cgirope;
	clockinfo_t	 ci;

	cgirope = cgi_rope_new();

	/* Decide which protocol we are on */
	if (strchr(rope_chars(cgirope), '&')) {
		clock_info_long(&ci);
	}
	else {
		clock_info_short(&ci);
	}

	clock_show(&ci);
}
