#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include "vector.h"
#include "cgi.h"
#include "smalloc.h"
#include "utils.h"
#include "rope.h"
#include "log.h"

#ifdef DPRINTF
#undef DPRINTF
#define DPRINTF(x) do { } while(0)
#endif

static bool
datafile_match_strings(char *haystack, char *needle)
{
	return !!strstr(haystack, needle);
}

/* haystack, needle */
static int
datafile_match_vectors( vector_t *datavec,
						vector_t *inputvec,
						int *fields)
{
	int	 i, j;
	int	 val = 0;
	int	 len;

	len = fields[0];

	for (i = 0; i < vector_length(inputvec); i++) {
		for (j = 0; j < len; j++) {
			if (!fields[(j + 1)]) continue;

			if (!vector_data(datavec, j)) {
				log("cgisys", PRI_ALERT, "No data field in vector at "
								"offset %d in '%s'",
								j, implode(datavec, "|"));
				continue;
			}
			if (datafile_match_strings(
					vector_data(datavec, j),
					vector_data(inputvec, i)))
				val++;

		}
	}

	return val;
}

static vector_t *
datafile_input_vector (char *str)
{
	vector_t	*inputvec;
	char		*cp;
	char		*sp;
	char		*base;

	DPRINTF((stderr, "Making input vector from '%s'\n", str));
	inputvec = vector_new(sfree, 0);

	base = str;
	sp = str;
	for (cp = str; *cp; cp++) {
		if (isupper(*cp))
			*sp++ = tolower(*cp);
		else if (islower(*cp))
			*sp++ = *cp;
		else if (isspace(*cp)) {
			*sp++ = '\0';
			if (strlen(base))
				vector_push_back(inputvec, sstrdup(base));
			DPRINTF((stderr, "Got string '%s'\n", base));
			base = sp;
		}
	}

	/* Grab last word */
	if (strlen(base))
		vector_push_back(inputvec, sstrdup(base));
	DPRINTF((stderr, "Got string '%s'\n", base));

	return inputvec;
}

static vector_t *
datafile_data_vector (char *line)
{
	return explode(line, "|");
}

/* Int array, first element is length, x[i+1] is a boolean */
static int *
datafile_field_array (vector_t *fieldvec, int max)
{
	int		*fields;
	int		 i, j;
	char	*data;

	fields = (int *)smalloc((max + 1) * sizeof(int));
	memset(fields, 0, (max + 1) * sizeof(int));
	fields[0] = max;

	for (i = 0; i < vector_length(fieldvec); i++) {
		data = vector_data(fieldvec, i);

		DPRINTF((stderr, "Got vector data %s\n", data));

		if (!data) {
			log("cgisys", PRI_CRIT, "Got NULL pointer in field "
							"vector at %d", i);
			exit(0);
		}

		if (*data == '*') {
			for (j = 0; j < max; j++)
				fields[(j + 1)] = TRUE;
			break;	/* Might as well */
		}

		j = atol(data);
		if ((j >= max) || (j < 0)) {
			log("cgisys", PRI_WARNING, "Got field number out of range: "
							"Max: %d, Requested: %d", max, j);
			continue;
		}
		fields[(j + 1)] = TRUE;
	}

#ifdef DEBUG
	fprintf(stderr, "Field count %d (from %d)\n", fields[0], max);
	for (i = 0; i < fields[0]; i++) {
		fprintf(stderr, "Field %d %sset\n", i, (fields[(i+1)]?"":"un"));
	}
#endif

	return fields;
}

static FILE *
datafile_stream (rope_t *pathstr)
{
	FILE	*input;
	char	*file;			/* Path to datafile */

	file = rope_chars(pathstr);

	if (access(file, R_OK) != 0) {
		log("cgiconf", PRI_CRIT, "Could not open datafile '%s' for "
						"reading: %s", file, strerror(errno));
		exit(0);
	}

	input = fopen(file, "r");
	if (!input) {
		log("cgiconf", PRI_CRIT, "Could not open datafile '%s' for "
						"reading: %s", file, strerror(errno));
		exit(0);
	}

	return input;
}

vector_t *
datafile_read(rope_t *pathstr, vector_t *fieldvec, char *search)
{
	int			*fields;		/* Field array, size in [0] */
	int			 value;
	char		 line[BUFSIZ];	/* Input line (fgets()) */
	FILE		*input;			/* Input FILE* */
	vector_t	*results;		/* Quadratic results vector */
	vector_t	*datavec;		/* Data vector */
	vector_t	*inputvec;		/* Input vector */

	input = datafile_stream(pathstr);				/* Get stream */
	inputvec = datafile_input_vector(search); /* Get words */
	results = vector_new((delete_t)vector_delete, 0);	/* Allocate */

/* Read first line to find count of fields */
	if (!fgets(line, BUFSIZ, input))
		return results;		/* No input lines */
	line[strcspn(line, "\r\n")] = '\0';
	datavec = datafile_data_vector(line);
	DPRINTF((stderr, "Line: %s\n", line));
	DPRINTF((stderr, "Implode %s\n", implode(datavec, "|")));
	DPRINTF((stderr, "Vector length %d\n", vector_length(datavec)));

/* Using this input line, generate a field set */
	fields = datafile_field_array(fieldvec, vector_length(datavec));

/* Jump back to start of file */
	fseek(input, 0, SEEK_SET);

/* Process the file */
	while (fgets(line, BUFSIZ, input)) {
		line[strcspn(line, "\r\n")] = '\0';
		datavec = datafile_data_vector(line);

		value = datafile_match_vectors(datavec, inputvec, fields);

		if (value)
			vector_push_back(results, datavec);
	}

	vector_delete(inputvec);
	sfree(fields);
	fclose(input);

	return results;
}
