/*
 * C utility library
 * Copyright (C) 1999 Open World Ltd
 * Written by Shevek <shevek@anarres.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "library.h"
#include "cgi.h"
#include "http.h"
#include "hash.h"
#include "smalloc.h"
#include "config.h"
#include "rope.h"
#include "log.h"

#define CGI_MATCH_HACK

char	*cgi_input;

void
cgi_error(char *str)
{
	log("cgisys", PRI_ERR, "(cgi_error: %s", str);
}

void
cgi_fatal(char *str)
{
	cgi_error(str);
	exit(1);
}

static void
cgi_init(void)
{
	char	*clenstr;
	char	*querystr;
	char	*method;
	long	 inputlen;

	method = getenv("REQUEST_METHOD");
	clenstr = getenv("CONTENT_LENGTH");
	if (!method) {
		if (clenstr)
			method="POST";
		else
			method="GET";
	}

	if (strcmp(method, "POST") == 0) {
		if (clenstr) {
			inputlen = atol(clenstr);
			cgi_input = (char *)smalloc((inputlen + 1) * sizeof(char));
			inputlen = fread(cgi_input, sizeof(char), inputlen, stdin);
			cgi_input[inputlen] = '\0';
		}
		else {
			log("cgisys", PRI_CRIT, "cgi_init: No CONTENT_LENGTH");
			exit(0);
		}
	}
	else if (strcmp(method, "GET") == 0) {
		querystr = getenv("QUERY_STRING");
		if (querystr) {
			cgi_input = strdup(querystr);
		}
		else {
			log("cgisys", PRI_CRIT, "cgi_init: No QUERY_STRING");
			exit(0);
		}
	}
}

/* Return a malloc()ed un-HTTP escaped copy of str from a URL */
char *
cgi_url_decode (char *str)
{
	char    *new;
	char    *cp;
	new = strdup(str);
	for (cp = new; *cp; cp++)
		if (*cp == '+') *cp = ' ';
	HTUnEscape(new);
	return new;
}

char *
cgi_url_encode (char *str)
{
	char    *new;
	char    *cp;

	new = HTEscape(str);
	for (cp = new; *cp; cp++)
		if (*cp == ' ') *cp = '+';
	return new;
}

rope_t *
cgi_rope_new(void)
{
	if (!cgi_input)	/* Static variable */
		cgi_init();
	return rope_new(cgi_input);
}

hash_t *
cgi_hash_new(void)
{
	char	*input;
	hash_t	*vars;
	char	*key;
	char	*value;

	if (!cgi_input)	/* Static variable */
		cgi_init();

	input = strdup(cgi_input);
	vars = hash_new(free, CGI_HASHSIZE);

	key = strtok(input, "&");
	// DPRINTF((stderr, "CGI input is %s\n", input));
	while (key) {
		// DPRINTF((stderr, "KEY is %s\n", key));
#ifdef CGI_MATCH_HACK
		if (strncmp(key, "Match", 5) == 0) {
			value = strchr(key, '=');
			if (value) {
				*value++ = '\0';
				/* malloc()s */
				value = cgi_url_decode(value);
				hash_add(vars, "Text", value);
			}
			value = strchr(key, '+');
			if (value) {
				*value = '=';
			}
			else {
				log("cgisys", PRI_WARNING, "cgi_hash_new: Expected a "
								"space in Match='%s' value", key);
			}
		}
#endif
		value = strchr(key, '=');
		if (value) {
			*value++ = '\0';
			/* malloc()s */
			value = cgi_url_decode(value);
			hash_add(vars, key, value);
		}
		else {
			log("cgisys", PRI_INFO, "No value for key '%s' in form",
							key);
		}
		key = strtok(NULL, "&");
	}

	return vars;
}

/* BEWARE: This uses static allocation */
char *
cgi_html_encode (char *str)
{
	static char	*new = NULL;
	static int	 length = 0;
	int			 newlength = 1;
	char		*cp;

	for (cp = str; *cp; cp++) {
		if (strchr("&", *cp)) newlength += 5;
		else if (strchr("<>\n", *cp)) newlength += 4;
		else newlength++;
	}

	if (newlength > length) {
		new = (char *)realloc(new, sizeof(char) * newlength);
		length = newlength;
	}

	cp = new;
	while (*str) {
		switch(*str) {
			case '&':
				strncpy(cp, "&amp;", 5);
				cp += 5;
				break;
			case '<':
				strncpy(cp, "&lt;", 4);
				cp += 4;
				break;
			case '>':
				strncpy(cp, "&gt;", 4);
				cp += 4;
				break;
			case '\n':
				strncpy(cp, "<BR>", 4);
				cp += 4;
				break;
			default:
				*cp++ = *str;
		}
		str++;
	}
	*cp = '\0';
	return new;
}
